from django.template import Template, Context
from django.test import TestCase

from ..models import Award
from ..templatetags.badgify_tags import badgify_badges as badges_tag

from .models import BadgifyUser as User
from .mixins import BadgeFixturesMixin, UserFixturesMixin


class BadgesTagTestCase(TestCase, BadgeFixturesMixin, UserFixturesMixin):
    """
    ``badgify_badges`` templatetag test case.
    """

    def setUp(self):
        self.badges = []
        badges, slugs = self.get_dummy_badges()
        for badge in badges:
            badge.save()
            self.badges.append(badge)
        self.users = self.create_users()
        self.badge = self.badges[0]
        self.user = User.objects.get(username='johndoe')
        self.award = Award.objects.create(badge=self.badge, user=self.user)

    def test_no_args(self):
        badges = badges_tag()
        self.assertEqual(len(badges), 20)
        template = Template('''
            {% load badgify_tags %}
            {% badgify_badges as badges %}
            {% for badge in badges %}{{ badge.name }} {{ badge.slug }}{% endfor %}
        ''')
        rendered = template.render(Context({}))
        for badge in self.badges:
            self.assertIn(badge.name, rendered)
            self.assertIn(badge.slug, rendered)

    def test_with_username(self):
        badges = badges_tag(**{'username': 'johndoe'})
        self.assertEqual(len(badges), 1)
        template = Template('''
            {% load badgify_tags %}
            {% badgify_badges username="johndoe" as badges %}
            {% for badge in badges %}{{ badge.name }} {{ badge.slug }}{% endfor %}
        ''')
        rendered = template.render(Context({}))
        self.assertIn(self.badge.name, rendered)
        self.assertIn(self.badge.slug, rendered)

    def test_with_user(self):
        badges = badges_tag(**{'user': self.user})
        self.assertEqual(len(badges), 1)
        template = Template('''
            {% load badgify_tags %}
            {% badgify_badges user=user as badges %}
            {% for badge in badges %}{{ badge.name }} {{ badge.slug }}{% endfor %}
        ''')
        rendered = template.render(Context({'user': self.user}))
        self.assertIn(self.badge.name, rendered)
        self.assertIn(self.badge.slug, rendered)



from django.db import models
from django.contrib.auth.models import AbstractUser


class BadgifyUser(AbstractUser):
    love_python = models.BooleanField(default=False)
    love_js = models.BooleanField(default=False)
    love_java = models.BooleanField(default=False)



# -*- coding: utf-8 -*-
import os


BASE_DIR = os.path.dirname(os.path.dirname(__file__))
PROJECT_DIR = os.path.dirname(__file__)

SECRET_KEY = 'l8h=t&zq%7f_vv8ge3@0jw^qgvedz*82j2$2rkbr^)i*y0p&x('

DEBUG = True

ALLOWED_HOSTS = []

INSTALLED_APPS = (
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'badgify',
    'example',
)

MIDDLEWARE_CLASSES = (
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.auth.middleware.SessionAuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
)

ROOT_URLCONF = 'example.urls'
WSGI_APPLICATION = 'example.wsgi.application'

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(PROJECT_DIR, 'db.sqlite3'),
    }
}

LANGUAGE_CODE = 'en-us'
TIME_ZONE = 'UTC'
USE_I18N = True
USE_L10N = True
USE_TZ = True

MEDIA_URL = '/media/'
MEDIA_ROOT = os.path.join(BASE_DIR, 'badgify', 'tests', 'media')
STATIC_URL = '/static/'
STATIC_ROOT = os.path.join(BASE_DIR, 'badgify', 'tests', 'static')

AUTH_USER_MODEL = 'example.User'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.contrib.auth.context_processors.auth',
                'django.template.context_processors.debug',
                'django.template.context_processors.i18n',
                'django.template.context_processors.media',
                'django.template.context_processors.static',
                'django.template.context_processors.tz',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'verbose': {
            'format': '%(levelname)s %(asctime)s %(message)s'
        },
    },
    'handlers': {
        'stream': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
            'formatter': 'verbose'
        },
    },
    'loggers': {
        'django.request': {
            'handlers': ['stream'],
            'level': 'DEBUG',
            'propagate': True,
        },
        'example.fixtures': {
            'handlers': ['stream'],
            'level': 'DEBUG',
            'propagate': True,
        },
        'badgify': {
            'handlers': ['stream'],
            'level': 'DEBUG',
            'propagate': True,
        },
    },
}



from django.urls import re_path as url

from .views import BadgeListView, BadgeDetailView


urlpatterns = [
    url(r'^$', BadgeListView.as_view(), name='badge_list'),
    url(r'^(?P<slug>[\w-]+)/$', BadgeDetailView.as_view(), name='badge_detail'),
]



# -*- coding: utf-8 -*-
from django.db import models
from django.contrib.auth.models import AbstractUser
from django.utils.translation import gettext_lazy as _


class User(AbstractUser):
    love_python = models.BooleanField(default=False)
    love_js = models.BooleanField(default=False)

    class Meta:
        verbose_name = _('user')
        verbose_name_plural = _('users')



import logging

from importlib import import_module

from django.core import exceptions
from django.db import connection

import six

from . import settings

logger = logging.getLogger(__name__)


CLASS_PATH_ERROR = 'django-badgify is unable to interpret settings value for %s. '\
                   '%s should be in the form of a tupple: '\
                   '(\'path.to.models.Class\', \'app_label\').'


def load_class(class_path, setting_name=None):
    """
    Loads a class given a class_path. The setting value may be a string or a
    tuple. The setting_name parameter is only there for pretty error output, and
    therefore is optional

    Borrowed from: https://github.com/thoas/django-discussions/blob/master/discussions/utils.py
    """
    if not isinstance(class_path, six.string_types):
        try:
            class_path, app_label = class_path
        except Exception:
            if setting_name:
                raise exceptions.ImproperlyConfigured(CLASS_PATH_ERROR % (
                    setting_name, setting_name))
            else:
                raise exceptions.ImproperlyConfigured(CLASS_PATH_ERROR % (
                    'this setting', 'It'))

    try:
        class_module, class_name = class_path.rsplit('.', 1)
    except ValueError:
        if setting_name:
            txt = '%s isn\'t a valid module. Check your %s setting' % (
                class_path, setting_name)
        else:
            txt = '%s isn\'t a valid module.' % class_path
        raise exceptions.ImproperlyConfigured(txt)

    try:
        mod = import_module(class_module)
    except ImportError as e:
        if setting_name:
            txt = 'Error importing backend %s: "%s". Check your %s setting' % (
                class_module, e, setting_name)
        else:
            txt = 'Error importing backend %s: "%s".' % (class_module, e)
        raise exceptions.ImproperlyConfigured(txt)

    try:
        clazz = getattr(mod, class_name)
    except AttributeError:
        if setting_name:
            txt = ('Backend module "%s" does not define a "%s" class. Check'
                   ' your %s setting' % (class_module, class_name,
                                         setting_name))
        else:
            txt = 'Backend module "%s" does not define a "%s" class.' % (
                class_module, class_name)
        raise exceptions.ImproperlyConfigured(txt)
    return clazz


def get_model_string(model_name):
    """
    Returns the model string notation Django uses for lazily loaded ForeignKeys
    (eg 'auth.User') to prevent circular imports.
    This is needed to allow our crazy custom model usage.

    Borrowed from: https://github.com/thoas/django-discussions/blob/master/discussions/utils.py
    """
    setting_name = 'BADGIFY_%s_MODEL' % model_name.upper().replace('_', '')
    class_path = getattr(settings, setting_name, None)
    if not class_path:
        return 'badgify.%s' % model_name
    elif isinstance(class_path, str):
        parts = class_path.split('.')
        try:
            index = parts.index('models') - 1
        except ValueError:
            raise exceptions.ImproperlyConfigured(CLASS_PATH_ERROR % (
                setting_name, setting_name))
        app_label, model_name = parts[index], parts[-1]
    else:
        try:
            class_path, app_label = class_path
            model_name = class_path.split('.')[-1]
        except Exception:
            raise exceptions.ImproperlyConfigured(CLASS_PATH_ERROR % (
                setting_name, setting_name))
    return '%s.%s' % (app_label, model_name)


def chunks(l, n):
    """
    Yields successive n-sized chunks from l.
    """
    for i in range(0, len(l), n):
        yield l[i:i + n]


def log_queries(recipe):
    """
    Logs recipe instance SQL queries (actually, only time).
    """
    logger.debug(
        '⚐ Badge %s: SQL queries time %.2f second(s)',
        recipe.slug,
        sum([float(q['time']) for q in connection.queries]))


def sanitize_command_options(options):
    """
    Sanitizes command options.
    """
    multiples = [
        'badges',
        'exclude_badges',
    ]

    for option in multiples:
        if options.get(option):
            value = options[option]
            if value:
                options[option] = [v for v in value.split(' ') if v]

    return options



from django.urls import reverse
from django.db import models
from django.utils.text import slugify

from badgify.compat import gettext_lazy as _
from badgify import compat, settings


class Badge(models.Model):
    """
    A Badge.
    """
    name = models.CharField(
        max_length=255,
        verbose_name=_('name'),
        help_text=_('The badge name'))

    slug = models.SlugField(
        max_length=255,
        blank=True,
        unique=True,
        verbose_name=_('slug'),
        help_text=_('The badge slug (auto-generated if empty)'))

    description = models.TextField(
        blank=True,
        verbose_name=_('description'),
        help_text=_('The badge description'))

    image = models.ImageField(null=True, blank=True,
                              verbose_name=_('Image'),
                              help_text=_('Please, upload an image for this badge'),
                              upload_to=settings.BADGE_IMAGE_UPLOAD_ROOT)

    users = models.ManyToManyField(
        compat.AUTH_USER_MODEL,
        through='Award',
        verbose_name=_('users'),
        help_text=_('Users that earned this badge'))

    users_count = models.IntegerField(
        default=0,
        verbose_name=_('users count'),
        editable=False)

    class Meta:
        abstract = True
        app_label = 'badgify'
        verbose_name = _('badge')
        verbose_name_plural = _('badges')

    def __str__(self):
        return self.name

    def get_absolute_url(self):
        return reverse('badge_detail', kwargs={'slug': self.slug})

    def save(self, *args, **kwargs):
        if not self.slug:
            self.slug = slugify(self.name)
        return super(Badge, self).save(*args, **kwargs)



from django.urls import reverse
from django.test import TestCase

from ..models import Badge

from .mixins import BadgeFixturesMixin


class BadgeDetailViewTestCase(TestCase):
    """
    BadgeDetail view test case.
    """

    def setUp(self):
        self.badge = Badge.objects.create(
            name='Djangonaut',
            slug='djangonaut',
            description='Django Developer',
            image=u'')
        self.badge_with_hyphen = Badge.objects.create(
            name='Super Djangonaut',
            slug='super-djangonaut',
            description='Super Django Developer',
            image=u'')

    def test_view(self):
        for badge in [self.badge, self.badge_with_hyphen]:
            res = self.client.get(badge.get_absolute_url())
            self.assertEqual(res.status_code, 200)
            self.assertTemplateUsed(res, 'badgify/badge_detail.html')

    def test_view_404(self):
        res = self.client.get(reverse('badge_detail', kwargs={'slug': 'foobar'}))
        self.assertEqual(res.status_code, 404)


class BadgeListViewTestCase(TestCase, BadgeFixturesMixin):
    """
    BadgeList view test case.
    """

    def setUp(self):
        self.badges = []
        badges, slugs = self.get_dummy_badges()
        for badge in badges:
            badge.save()
            self.badges.append(badge)

    def test_view(self):
        res = self.client.get(reverse('badge_list'))
        self.assertEqual(res.status_code, 200)
        self.assertTemplateUsed(res, 'badgify/badge_list.html')



from django.test import TestCase

from ..exceptions import BadgeNotFound
from ..recipe import BaseRecipe
from ..registry import BadgifyRegistry as Registry

from .recipes import (
    Recipe1,
    Recipe2,
    BadRecipe)


class RegistryTestCase(TestCase):
    """
    Registry test case.
    """

    def test_recipes(self):
        registry = Registry()
        registry.register(Recipe1)
        self.assertTrue(isinstance(registry.recipes, dict))
        self.assertEqual(len(registry.recipes), 1)

    def test_registered(self):
        registry = Registry()
        registry.register(Recipe1)
        registry.register(Recipe2)
        self.assertTrue(isinstance(registry.registered, list))
        self.assertIn('recipe1', registry.registered)
        self.assertIn('recipe2', registry.registered)

    def test_register(self):
        registry = Registry()
        # With a single class
        self.assertRaises(AssertionError, registry.register, BadRecipe)
        self.assertIsNone(registry.register(Recipe1))
        self.assertEqual(len(registry.recipes), 1)
        # With a list of classes
        registry.clear()
        registry.register([Recipe1, Recipe2])
        self.assertEqual(len(registry.recipes), 2)

    def test_unregister(self):
        registry = Registry()
        registry.register(Recipe1)
        self.assertEqual(len(registry.recipes), 1)
        registry.unregister(Recipe1)
        self.assertEqual(len(registry.recipes), 0)

    def test_clear(self):
        registry = Registry()
        registry.register(Recipe1)
        self.assertEqual(len(registry.recipes), 1)
        registry.clear()
        self.assertTrue(isinstance(registry.recipes, dict))
        self.assertEqual(len(registry.recipes), 0)

    def test_get_recipe_instance(self):
        registry = Registry()
        registry.register(Recipe1)
        instance = registry.get_recipe_instance('recipe1')
        self.assertTrue(isinstance(instance, BaseRecipe))
        self.assertRaises(BadgeNotFound, registry.get_recipe_instance, 'unknown')

    def test_get_recipe_instances(self):
        registry = Registry()
        registry.register([Recipe1, Recipe2])
        # Without passing badges arg
        instances = registry.get_recipe_instances()
        self.assertEqual(len(instances), 2)
        for instance in instances:
            self.assertTrue(isinstance(instance, BaseRecipe))
        # By passing badges arg
        registry = Registry()
        registry.register([Recipe1, Recipe2])
        instances = registry.get_recipe_instances(badges=['recipe1', 'oops'])
        self.assertEqual(len(instances), 1)
        instances = registry.get_recipe_instances(badges=['recipe1', 'recipe2'])
        self.assertEqual(len(instances), 2)
        # By passing excluded arg
        registry = Registry()
        registry.register([Recipe1, Recipe2])
        instances = registry.get_recipe_instances(excluded=['recipe2'])
        self.assertEqual(len(instances), 1)
        self.assertTrue(isinstance(instances[0], Recipe1))

    def test_get_recipe_instances_for_badges(self):
        registry = Registry()
        registry.register([Recipe1, Recipe2])
        valid, invalid = registry.get_recipe_instances_for_badges(badges=['recipe1', 'recipe2', 'oops'])
        self.assertEqual(len(valid), 2)
        self.assertEqual(len(invalid), 1)
        self.assertIn('oops', invalid)
        # Test by passing a string instead of a list
        registry = Registry()
        registry.register(Recipe1)
        valid, invalid = registry.get_recipe_instances_for_badges(badges='recipe1')
        self.assertEqual(len(valid), 1)
        self.assertEqual(len(invalid), 0)

    def test_get_recipe_instance_from_class(self):
        registry = Registry()
        instance = registry.get_recipe_instance_from_class(Recipe1)
        self.assertTrue(isinstance(instance, Recipe1))
        self.assertRaises(AssertionError, registry.get_recipe_instance_from_class, BadRecipe)



import logging

from django.dispatch import receiver
from django.db.models.signals import post_save, pre_delete

from .models import Award

logger = logging.getLogger('badgify')


@receiver(post_save, sender=Award, dispatch_uid='badgify.award.post_save.increment_badge_users_count')
def increment_badge_users_count(sender, instance, created, **kwargs):
    from django.db.models import F
    from .settings import AUTO_DENORMALIZE

    if created and AUTO_DENORMALIZE:
        instance.badge.users_count = F('users_count') + 1
        instance.badge.save()


@receiver(pre_delete, sender=Award, dispatch_uid='badgify.award.pre_delete.decrement_badge_users_count')
def decrement_badge_users_count(sender, instance, **kwargs):
    from django.db.models import F
    from .settings import AUTO_DENORMALIZE

    if AUTO_DENORMALIZE and instance.badge.users_count >= 1:
        instance.badge.users_count = F('users_count') - 1
        instance.badge.save()



from imp import reload

from django.db.models import signals
from django.test import TestCase

from .. import settings
from .. import commands
from .. import registry
from ..models import Badge, Award
from ..compat import get_user_model

from .recipes import Recipe1, Recipe2


class CommandsTestCase(TestCase):
    """
    Commands test case.
    """

    def setUp(self):
        reload(settings)
        registry.clear()

        self.post_save_count = 0
        self.pre_delete_count = 0

        signals.post_save.connect(self._award_post_save_receiver, sender=Award)
        signals.pre_delete.connect(self._award_pre_delete_receiver, sender=Award)

    def tearDown(self):
        registry.clear()

        signals.post_save.disconnect(self._award_post_save_receiver, sender=Award)
        signals.pre_delete.disconnect(self._award_pre_delete_receiver, sender=Award)

    def _award_post_save_receiver(self, *args, **kwargs):
        self.post_save_count += 1

    def _award_pre_delete_receiver(self, *args, **kwargs):
        self.pre_delete_count += 1

    def test_sync_badges(self):
        registry.register([Recipe1, Recipe2])
        created = commands.sync_badges()
        self.assertEqual(len(created), 2)

    def test_sync_badges_update(self):
        registry.register([Recipe1, Recipe2])
        commands.sync_badges()

        Recipe1.name = 'Howdy'
        commands.sync_badges(update=True)

        self.assertEqual(Badge.objects.get(slug='recipe1').name, 'Howdy')
        Recipe1.name = 'Recipe 1'

    def test_sync_count(self):
        settings.AUTO_DENORMALIZE = False

        user = get_user_model().objects.create_user('user', 'user@example.com', '$ecret')
        registry.register([Recipe1, Recipe2])
        commands.sync_badges()
        updated, unchanged = commands.sync_counts()

        self.assertEqual(len(updated), 0)
        self.assertEqual(len(unchanged), 2)

        Award.objects.create(user=user, badge=Badge.objects.get(slug='recipe1'))
        updated, unchanged = commands.sync_counts()

        self.assertEqual(len(updated), 1)
        self.assertEqual(len(unchanged), 1)

    def test_sync_awards_auto_denormalize_false(self):
        settings.AUTO_DENORMALIZE = False

        user = get_user_model().objects.create_user('user', 'user@example.com', '$ecret')
        registry.register(Recipe1)
        created = commands.sync_badges()
        recipe = registry.get_recipe_instance('recipe1')

        self.assertEqual(len(created), 1)
        self.assertEqual(recipe.badge.users.count(), 0)

        user.love_python = True
        user.save()

        commands.sync_awards()

        self.assertEqual(recipe.badge.users.count(), 1)
        self.assertEqual(recipe.badge.users_count, 0)
        self.assertEqual(self.post_save_count, 1)

    def test_sync_awards_auto_denormalize_true(self):
        settings.AUTO_DENORMALIZE = True

        user = get_user_model().objects.create_user('user1', 'user1@example.com', '$ecret')
        registry.register(Recipe1)
        commands.sync_badges()
        recipe = registry.get_recipe_instance('recipe1')

        user.love_python = True
        user.save()

        commands.sync_awards()

        self.assertEqual(recipe.badge.users.count(), 1)
        self.assertEqual(recipe.badge.users_count, 1)
        self.assertEqual(self.post_save_count, 1)
        self.assertEqual(self.pre_delete_count, 0)

        Award.objects.all().delete()

        self.assertEqual(self.pre_delete_count, 1)
        self.assertEqual(recipe.badge.users.count(), 0)
        self.assertEqual(recipe.badge.users_count, 0)

    def test_sync_awards_disable_signals(self):
        settings.AUTO_DENORMALIZE = True

        user = get_user_model().objects.create_user('user', 'user@example.com', '$ecret')
        registry.register(Recipe1)
        created = commands.sync_badges()
        recipe = registry.get_recipe_instance('recipe1')

        self.assertEqual(len(created), 1)
        self.assertEqual(self.post_save_count, 0)
        self.assertEqual(recipe.badge.users_count, 0)

        user.love_python = True
        user.save()

        commands.sync_awards(**{'disable_signals': True})

        self.assertEqual(recipe.badge.users_count, 0)
        self.assertEqual(self.post_save_count, 0)



# -*- coding: utf-8 -*-
from django.contrib.staticfiles.storage import staticfiles_storage

import badgify
from ..recipe import BaseRecipe
from .models import BadgifyUser as User


class PythonLoverRecipe(BaseRecipe):
    """
    People loving Python.
    """
    name = 'Python Lover'
    slug = 'python-lover'
    description = 'People loving Python programming language'

    @property
    def image(self):
        return staticfiles_storage.open('python-lover.png')

    @property
    def user_ids(self):
        return User.objects.filter(love_python=True).values_list('id', flat=True)


class JSLoverRecipe(BaseRecipe):
    """
    People loving JS.
    """
    name = 'JS Lover'
    slug = 'js-lover'
    description = 'People loving JS programming language'

    @property
    def image(self):
        return staticfiles_storage.open('js-lover.png')

    @property
    def user_ids(self):
        return User.objects.filter(love_js=True).values_list('id', flat=True)


class JavaLoverRecipe(BaseRecipe):
    """
    People loving Java.
    """
    name = 'Java Lover'
    slug = 'java-lover'
    description = 'People loving Java programming language'

    @property
    def image(self):
        return staticfiles_storage.open('js-lover.png')


badgify.register([
    PythonLoverRecipe,
    JSLoverRecipe,
    JavaLoverRecipe,
])



# -*- coding: utf-8 -*-
from django.core.management.base import CommandError, LabelCommand

from badgify import commands
from badgify.utils import sanitize_command_options


class Command(LabelCommand):
    """
    Command that synchronizes badges, awards and counts.
    """
    help = 'Synchronizes badges, awards and counts'

    def add_arguments(self, parser):
        """
        Command arguments.
        """
        super(Command, self).add_arguments(parser)

        parser.add_argument('--badges',
                            action='store',
                            dest='badges',
                            type=str)

        parser.add_argument('--db-read',
                            action='store',
                            dest='db_read',
                            type=str)

        parser.add_argument('--disable-signals',
                            action='store_true',
                            dest='disable_signals')

        parser.add_argument('--batch-size',
                            action='store',
                            dest='batch_size',
                            type=int)

        parser.add_argument('--update',
                            action='store_true',
                            dest='update')

        parser.add_argument('--exclude-badges',
                            action='store',
                            dest='exclude_badges',
                            type=str)

    def handle_label(self, label, **options):
        """
        Command handler.
        """
        if not hasattr(commands, 'sync_%s' % label):
            raise CommandError('"%s" is not a valid command.' % label)

        getattr(commands, 'sync_%s' % label)(**sanitize_command_options(options))



# -*- coding: utf-8 -*-
from django.core.management.base import BaseCommand

from badgify.commands import show_stats


class Command(BaseCommand):
    """
    Commands that shows badge stats.
    """
    help = 'Shows badge stats'

    def add_arguments(self, parser):
        """
        Command arguments.
        """
        parser.add_argument('--db-read',
                            action='store',
                            dest='db_read',
                            type=str)

    def handle(self, **options):
        """
        Command handler.
        """
        show_stats(**options)



from django.db import IntegrityError
from django.test import TestCase

from ..models import Badge, Award
from .mixins import UserFixturesMixin


class BadgeTestCase(TestCase):
    """
    Badge model test case.
    """

    def test_autocreate_slug(self):
        badge = Badge.objects.create(name='Super Chouette')
        self.assertEqual(badge.slug, 'super-chouette')


class AwardTestCase(TestCase, UserFixturesMixin):
    """
    Award model test case.
    """

    def setUp(self):
        self.create_users()

    def test_create(self):
        badge = Badge.objects.create(name='Super Chouette')
        Award.objects.create(user=self.user1, badge=badge)
        Award.objects.create(user=self.user2, badge=badge)
        self.assertEqual(badge.users.count(), 2)
        self.assertRaises(IntegrityError, Award.objects.create, **{
            'user': self.user1,
            'badge': badge
        })



from django.shortcuts import get_object_or_404
from django.utils.functional import cached_property
from django.views.generic import ListView

from .models import Badge
from . import settings


class BadgeListView(ListView):
    """
    Badge list view.
    """
    model = Badge
    context_object_name = 'badges'
    template_name = 'badgify/badge_list.html'
    paginate_by = settings.BADGE_LIST_VIEW_PAGINATE_BY


class BadgeDetailView(ListView):
    """
    Badge detail view.
    """
    model = Badge
    template_name = 'badgify/badge_detail.html'
    context_object_name = 'awarded_users'
    paginate_by = settings.BADGE_DETAIL_VIEW_PAGINATE_BY

    @cached_property
    def badge(self):
        return get_object_or_404(self.model, slug=self.kwargs.get('slug', None))

    def get_queryset(self):
        return self.badge.users.all()

    def get_context_data(self, **kwargs):
        context = super(BadgeDetailView, self).get_context_data(**kwargs)
        context['badge'] = self.badge
        return context



# Generated by Django 3.0.1 on 2019-12-30 02:55

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Award',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('awarded_at', models.DateTimeField(auto_now_add=True, verbose_name='awarded at')),
            ],
            options={
                'verbose_name': 'award',
                'verbose_name_plural': 'awards',
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Badge',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(help_text='The badge name', max_length=255, verbose_name='name')),
                ('slug', models.SlugField(blank=True, help_text='The badge slug (auto-generated if empty)', max_length=255, unique=True, verbose_name='slug')),
                ('description', models.TextField(blank=True, help_text='The badge description', verbose_name='description')),
                ('image', models.ImageField(blank=True, help_text='Please, upload an image for this badge', null=True, upload_to='badges', verbose_name='Image')),
                ('users_count', models.IntegerField(default=0, editable=False, verbose_name='users count')),
                ('users', models.ManyToManyField(help_text='Users that earned this badge', through='badgify.Award', to=settings.AUTH_USER_MODEL, verbose_name='users')),
            ],
            options={
                'verbose_name': 'badge',
                'verbose_name_plural': 'badges',
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='award',
            name='badge',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='awards', to='badgify.Badge', verbose_name='badge'),
        ),
        migrations.AddField(
            model_name='award',
            name='user',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='badges', to=settings.AUTH_USER_MODEL, verbose_name='user'),
        ),
        migrations.AlterUniqueTogether(
            name='award',
            unique_together={('user', 'badge')},
        ),
    ]



# -*- coding: utf-8 -*-
from django import template

from ..models import Badge, Award
from ..compat import get_user_model


register = template.Library()
register_tag = register.assignment_tag if hasattr(register, 'assignment_tag') else register.simple_tag


@register_tag
def badgify_badges(**kwargs):
    """
    Returns all badges or only awarded badges for the given user.
    """
    User = get_user_model()
    user = kwargs.get('user', None)
    username = kwargs.get('username', None)
    if username:
        try:
            user = User.objects.get(username=username)
        except User.DoesNotExist:
            pass
    if user:
        awards = Award.objects.filter(user=user).select_related('badge')
        badges = [award.badge for award in awards]
        return badges
    return Badge.objects.all()



import django
from django.conf import settings

__all__ = ['get_user_model']


AUTH_USER_MODEL = getattr(settings, 'AUTH_USER_MODEL', 'auth.User')


def get_user_model():
    if django.VERSION >= (1, 5):
        from django.contrib.auth import get_user_model as _get_user_model
        User = _get_user_model()
    else:
        from django.contrib.auth.models import User
    return User


try:
    from django.utils.translation import gettext_lazy as gettext_lazy
except ImportError:
    from django.utils.translation import gettext_lazy  # noqa



from random import randrange

from django.utils.text import slugify

from random_words import RandomNicknames

from ..recipe import BaseRecipe
from ..models import Badge

from .models import BadgifyUser as User


class UserFixturesMixin(object):
    """
    User fixtures mixin.
    """

    def create_users(self):
        self.user1 = User.objects.create_user(
            'johndoe',
            'john@example.com',
            'secret')
        self.user2 = User.objects.create_user(
            'mikedavis',
            'mike@example.com',
            'secret')
        self.user3 = User.objects.create_user(
            'banana',
            'banana@example.com',
            'secret')

    def get_dummy_users(self, count=20):
        rn = RandomNicknames()
        users, usernames = [], []
        for name in rn.random_nicks(count=count):
            name = '%s #%d' % (name, randrange(99))
            username = slugify(name)
            usernames.append(username)
            users.append(User(
                username=username,
                email='%s@example.com' % username,
                password='secret'))
        return (users, usernames)


class BadgeFixturesMixin(object):
    """
    Badge fixtures mixin.
    """

    def get_dummy_badges(self, count=20):
        rn = RandomNicknames()
        badges, slugs = [], []
        for name in rn.random_nicks(count=count):
            name = '%s #%d' % (name, randrange(99))
            slug = slugify(name)
            slugs.append(slug)
            badges.append(Badge(name=name, slug=slug, description=name))
        return badges, slugs


class RecipeFixturesMixin(object):
    """
    Recipe fixtures mixin.
    """

    def get_dummy_recipes(self, count=50):
        rn = RandomNicknames()
        recipes, slugs = [], []
        for name in rn.random_nicks(count=count):
            slug = slugify(name)
            slugs.append(slug)
            name = name.encode('ascii', 'ignore')
            attrs = dict(
                name=name,
                slug=slug,
                description=name,
                image=name,
                user_queryset=User.objects.none())
            recipes.append(type(name, (BaseRecipe,), attrs))
        return (recipes, slugs)



import logging

from django.db import DEFAULT_DB_ALIAS, IntegrityError
from django.db.models import signals
from django.utils.functional import cached_property

from . import settings
from .compat import get_user_model
from .models import Badge, Award
from .utils import chunks

logger = logging.getLogger('badgify')

PRE_DELETE_UID = 'badgify.award.pre_delete.decrement_badge_users_count'


class BaseRecipe(object):
    """
    Base class for badge recipes.
    """

    # Badge.name
    name = None

    # Badge.slug
    slug = None

    # Badge.description
    description = None

    # The database on which to perform read queries
    db_read = DEFAULT_DB_ALIAS

    # How many awards to create at once
    batch_size = settings.BATCH_SIZE

    @property
    def image(self):
        raise NotImplementedError('Image must be implemented')

    @property
    def user_ids(self):
        pass

    @property
    def badge(self):
        return self.cached_badge or self.uncached_badge

    @cached_property
    def cached_badge(self):
        return self.get_badge()

    @property
    def uncached_badge(self):
        return self.get_badge()

    def get_badge(self):
        """
        The related ``Badge`` object.
        """
        try:
            obj = Badge.objects.using(self.db_read).get(slug=self.slug)
            logger.debug('✓ Badge %s: fetched from db (%s)', obj.slug, self.db_read)
        except Badge.DoesNotExist:
            obj = None
        return obj

    def create_badge(self, update=False):
        """
        Saves the badge in the database (or updates it if ``update`` is ``True``).
        Returns a tuple: ``badge`` (the badge object) and ``created`` (``True``, if
        badge has been created).
        """
        badge, created = self.badge, False
        if badge:
            logger.debug('✓ Badge %s: already created', badge.slug)
            if update:
                to_update = {}
                for field in ('name', 'slug', 'description', 'image'):
                    attr = getattr(self, field)
                    badge_attr = getattr(badge, field)
                    if attr != badge_attr:
                        to_update[field] = attr
                        logger.debug('✓ Badge %s: updated "%s" field', self.slug, field)
                Badge.objects.filter(id=badge.id).update(**to_update)
        else:
            kwargs = {'name': self.name, 'image': self.image}
            optional_fields = ['slug', 'description']
            for field in optional_fields:
                value = getattr(self, field)
                if value is not None:
                    kwargs[field] = value
            badge = Badge.objects.create(**kwargs)
            created = True
            logger.debug('✓ Badge %s: created', badge.slug)
        return (badge, created)

    def can_perform_awarding(self):
        """
        Checks if we can perform awarding process (is ``user_ids`` property
        defined? Does Badge object exists? and so on). If we can perform db
        operations safely, returns ``True``. Otherwise, ``False``.
        """
        if not self.user_ids:
            logger.debug(
                '✘ Badge %s: no users to check (empty user_ids property)',
                self.slug)
            return False

        if not self.badge:
            logger.debug(
                '✘ Badge %s: does not exist in the database (run badgify_sync badges)',
                self.slug)
            return False

        return True

    def update_badge_users_count(self):
        """
        Denormalizes ``Badge.users.count()`` into ``Bagdes.users_count`` field.
        """
        logger.debug('→ Badge %s: syncing users count...', self.slug)

        badge, updated = self.badge, False

        if not badge:
            logger.debug(
                '✘ Badge %s: does not exist in the database (run badgify_sync badges)',
                self.slug)
            return (self.slug, updated)

        old_value, new_value = badge.users_count, badge.users.count()

        if old_value != new_value:
            badge = Badge.objects.get(slug=self.slug)
            badge.users_count = new_value
            badge.save()
            updated = True

        if updated:
            logger.debug('✓ Badge %s: updated users count (from %d to %d)',
                         self.slug,
                         old_value,
                         new_value)
        else:
            logger.debug('✓ Badge %s: users count up-to-date (%d)',
                         self.slug,
                         new_value)

        return (badge, updated)

    def get_already_awarded_user_ids(self, db_read=None, show_log=True):
        """
        Returns already awarded user ids and the count.
        """

        db_read = db_read or self.db_read

        already_awarded_ids = self.badge.users.using(db_read).values_list('id', flat=True)
        already_awarded_ids_count = len(already_awarded_ids)

        if show_log:
            logger.debug(
                "→ Badge %s: %d users already awarded (fetched from db '%s')",
                self.slug,
                already_awarded_ids_count,
                db_read)

        return already_awarded_ids

    def get_current_user_ids(self, db_read=None):
        """
        Returns current user ids and the count.
        """
        db_read = db_read or self.db_read

        return self.user_ids.using(db_read)

    def get_unawarded_user_ids(self, db_read=None):
        """
        Returns unawarded user ids (need to be saved) and the count.
        """
        db_read = db_read or self.db_read

        already_awarded_ids = self.get_already_awarded_user_ids(db_read=db_read)
        current_ids = self.get_current_user_ids(db_read=db_read)
        unawarded_ids = list(set(current_ids) - set(already_awarded_ids))
        unawarded_ids_count = len(unawarded_ids)

        logger.debug(
            '→ Badge %s: %d users need to be awarded',
            self.slug,
            unawarded_ids_count)

        return (unawarded_ids, unawarded_ids_count)

    def get_obsolete_user_ids(self, db_read=None):
        """
        Returns obsolete users IDs to unaward.
        """
        db_read = db_read or self.db_read

        already_awarded_ids = self.get_already_awarded_user_ids(db_read=db_read, show_log=False)
        current_ids = self.get_current_user_ids(db_read=db_read)
        obsolete_ids = list(set(already_awarded_ids) - set(current_ids))
        obsolete_ids_count = len(obsolete_ids)

        logger.debug(
            '→ Badge %s: %d users need to be unawarded',
            self.slug,
            obsolete_ids_count)

        return (obsolete_ids, obsolete_ids_count)

    def create_awards(self, db_read=None, batch_size=None,
                      post_save_signal=True):
        """
        Create awards.
        """
        if not self.can_perform_awarding():
            return

        User = get_user_model()

        db_read = db_read or self.db_read
        batch_size = batch_size or self.batch_size

        unawarded_ids, unawarded_ids_count = self.get_unawarded_user_ids(db_read=db_read)
        obsolete_ids, obsolete_ids_count = self.get_obsolete_user_ids(db_read=db_read)

        if obsolete_ids:
            for user_ids in chunks(obsolete_ids, batch_size):
                obsolete_users = User.objects.using(db_read).in_bulk(user_ids).values()

                signals.pre_delete.disconnect(sender=Award, dispatch_uid=PRE_DELETE_UID)
                Award.objects.filter(user__in=user_ids).delete()

                logger.debug("→ Badge %s (db_read: %s): unawarded %s",
                             self.slug,
                             db_read,
                             ' '.join(['%s' % user for user in obsolete_users]))

        if unawarded_ids:
            for user_ids in chunks(unawarded_ids, batch_size):
                unwarded_users = User.objects.using(db_read).in_bulk(user_ids).values()
                objects = [Award(user_id=user_id, badge=self.badge) for user_id in user_ids]

                bulk_create_awards(objects=objects,
                                   batch_size=batch_size,
                                   post_save_signal=post_save_signal)

                logger.debug("→ Badge %s (db_read: %s): awarded %s",
                             self.slug,
                             db_read,
                             ' '.join(['%s' % user for user in unwarded_users]))


def bulk_create_awards(objects, batch_size=500, post_save_signal=True):
    """
    Saves award objects.
    """
    count = len(objects)
    if not count:
        return
    badge = objects[0].badge
    try:
        Award.objects.bulk_create(objects, batch_size=batch_size)
        if post_save_signal:
            for obj in objects:
                signals.post_save.send(sender=obj.__class__, instance=obj, created=True)
    except IntegrityError:
        logger.error('✘ Badge %s: IntegrityError for %d awards', badge.slug, count)



from django.contrib import admin
from django.urls import reverse
from django.utils.html import format_html

from badgify.compat import gettext_lazy as _

from .models import Badge, Award
from . import settings


class BadgeAdmin(admin.ModelAdmin):
    """
    Badge model admin options.
    """
    list_display = ('name', 'description', 'link', 'image_thumbnail')
    prepopulated_fields = {'slug': ('name',)}
    search_fields = ('name', 'slug', 'description')

    def image_thumbnail(self, obj):
        if obj.image:
            return format_html('<img src="{0}" width="80">', obj.image.url)
        return '<img src="http://placehold.it/80x80">'
    image_thumbnail.short_description = _('image')
    image_thumbnail.allow_tags = True

    def link(self, obj):
        return format_html('<a href="{0}">{1}</a>',
                           self.view_on_site(obj),
                           self.view_on_site(obj))
    link.short_description = _('link')
    link.allow_tags = True

    def view_on_site(self, obj):
        return reverse('badge_detail', kwargs={'slug': obj.slug})


class AwardAdmin(admin.ModelAdmin):
    """
    Award model admin options.
    """
    list_display = ('user', 'badge', 'awarded_at')
    date_hierarchy = 'awarded_at'
    list_filter = ('badge',)
    search_fields = (
        'user__username',
        'user__first_name',
        'user__last_name',
        'user__email',
        'badge__name',
        'badge__slug',
        'badge__description')


if settings.REGISTER_ADMIN:
    admin.site.register(Badge, BadgeAdmin)
    admin.site.register(Award, AwardAdmin)



from django.conf import settings


APP_NAMESPACE = 'BADGIFY'

REGISTER_ADMIN = getattr(
    settings,
    '%s_REGISTER_ADMIN' % APP_NAMESPACE,
    True)


BADGE_IMAGE_UPLOAD_ROOT = getattr(
    settings,
    '%s_IMAGE_UPLOAD_ROOT' % APP_NAMESPACE,
    'badges')

BADGE_LIST_VIEW_PAGINATE_BY = getattr(
    settings,
    '%s_BADGE_LIST_VIEW_PAGINATE_BY' % APP_NAMESPACE,
    20)

BADGE_DETAIL_VIEW_PAGINATE_BY = getattr(
    settings,
    '%s_BADGE_DETAIL_VIEW_PAGINATE_BY' % APP_NAMESPACE,
    20)

BADGE_MODEL = getattr(
    settings,
    '%s_BADGE_MODEL' % APP_NAMESPACE,
    'badgify.models.badge.Badge')

AWARD_MODEL = getattr(
    settings,
    '%s_AWARD_MODEL' % APP_NAMESPACE,
    'badgify.models.award.Award')

BATCH_SIZE = getattr(
    settings,
    '%s_BATCH_SIZE' % APP_NAMESPACE,
    500)

AUTO_DENORMALIZE = getattr(
    settings,
    '%s_AUTO_DENORMALIZE' % APP_NAMESPACE,
    True)



# -*- coding: utf-8 -*-
import logging
import random

from django.core import management
from django.core.management.base import BaseCommand
from django.db import DEFAULT_DB_ALIAS, IntegrityError
from django.utils.text import slugify

from random_words import RandomNicknames

from badgify.models import Badge, Award

from example.models import User

logger = logging.getLogger('example.fixtures')


class Command(BaseCommand):
    """
    Creates fixtures.
    """
    help = u'Create fixtures'

    def add_arguments(self, parser):
        """
        Command arguments.
        """
        parser.add_argument('--flushdb',
                            action='store_true',
                            dest='flushdb')

    def handle(self, **options):
        """
        Command handler.
        """
        self.flushdb = options.get('flushdb')
        self._pre_tasks()
        self._create_users()
        self._create_badges()
        self._create_awards()

    def _pre_tasks(self):
        """
        Pre-tasks handler.
        """
        if self.flushdb:
            management.call_command('flush', verbosity=0, interactive=False)
            logger.info('Flushed database')

    def _create_users(self):
        """
        Creates users.
        """
        rn = RandomNicknames()

        for name in rn.random_nicks(count=50):
            username = '%s%d' % (slugify(name), random.randrange(1, 99))
            user = User.objects.create_user(
                username=username,
                email='%s@example.com' % username,
                password='secret')
            logger.info('Created user: %s', user.username)

    def _create_badges(self):
        """
        Creates badges.
        """
        rn = RandomNicknames()

        for name in rn.random_nicks(count=20):
            slug = slugify(name)
            badge = Badge.objects.create(
                name=name,
                slug=slug,
                description='Lorem ipsum dolor sit amet, consectetur adipisicing elit')
            logger.info('Created badge: %s', badge.name)

    def _create_awards(self):
        """
        Creates awards.
        """
        users = User.objects.all()

        for user in users:
            everyone_badge = Badge.objects.last()
            badge = Badge.objects.order_by('?')[0]
            try:
                award = Award.objects.create(user=user, badge=badge)
                everyone_award = Award.objects.create(user=user, badge=everyone_badge)
                logger.info('%s --- %s', award, everyone_award)
            except IntegrityError:
                pass



from ..recipe import BaseRecipe
from ..compat import get_user_model


class BadRecipe(object):
    pass


class NotImplementedRecipe(BaseRecipe):
    pass


class Recipe1(BaseRecipe):
    name = 'Recipe 1'
    slug = 'recipe1'
    description = 'Recipe 1 description'

    @property
    def image(self):
        return 'image'

    @property
    def user_ids(self):
        return (get_user_model().objects.filter(love_python=True)
                                .values_list('id', flat=True))


class Recipe2(BaseRecipe):
    name = 'Recipe 2'
    slug = 'recipe2'
    description = 'Recipe 2 description'

    @property
    def image(self):
        return 'image'

    @property
    def user_ids(self):
        return []



# -*- coding: utf-8 -*-
from django.conf import settings
from django.conf.urls import include, url
from django.conf.urls.static import static
from django.contrib import admin
from  django.views.generic import TemplateView

urlpatterns = [
    url(r'^admin/', include(admin.site.urls)),
    url(r'', include('badgify.urls')),
] + static(settings.STATIC_URL, document_root=settings.STATIC_ROOT)



# -*- coding: utf-8 -*-
import os
from setuptools import setup, find_packages


ROOT = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(ROOT, "README.rst")) as f:
    README = f.read()

setup(
    name="django-badgify",
    version="0.6.0",
    description="A reusable application to create your own badge engine using Django",
    long_description=README,
    author="Gilles Fabio",
    author_email="gilles.fabio@gmail.com",
    url="http://github.com/ulule/django-badgify",
    packages=find_packages(),
    zip_safe=False,
    include_package_data=True,
    install_requires=[
        "Pillow",
        "pytz",
        "six",
    ],
    tests_require=["coverage", "RandomWords"],
    classifiers=[
        "Environment :: Web Environment",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Utilities",
    ],
)



import logging

from django.db import reset_queries, DEFAULT_DB_ALIAS
from django.db.models import Count, signals

from . import registry
from . import settings
from .models import Badge, Award
from .utils import log_queries

logger = logging.getLogger('badgify')


def sync_badges(**kwargs):
    """
    Iterates over registered recipes and creates missing badges.
    """
    update = kwargs.get('update', False)
    created_badges = []
    instances = registry.get_recipe_instances()

    for instance in instances:
        reset_queries()
        badge, created = instance.create_badge(update=update)
        if created:
            created_badges.append(badge)
        log_queries(instance)

    return created_badges


def sync_counts(**kwargs):
    """
    Iterates over registered recipes and denormalizes ``Badge.users.count()``
    into ``Badge.users_count`` field.
    """
    badges = kwargs.get('badges')
    excluded = kwargs.get('exclude_badges')

    instances = registry.get_recipe_instances(badges=badges, excluded=excluded)
    updated_badges, unchanged_badges = [], []

    for instance in instances:
        reset_queries()
        badge, updated = instance.update_badge_users_count()
        if updated:
            updated_badges.append(badge)
        else:
            unchanged_badges.append(badge)
        log_queries(instance)

    return (updated_badges, unchanged_badges)


def sync_awards(**kwargs):
    """
    Iterates over registered recipes and possibly creates awards.
    """
    badges = kwargs.get('badges')
    excluded = kwargs.get('exclude_badges')
    disable_signals = kwargs.get('disable_signals')
    batch_size = kwargs.get('batch_size', None)
    db_read = kwargs.get('db_read', None)

    award_post_save = True

    if disable_signals:
        settings.AUTO_DENORMALIZE = False
        award_post_save = False

    instances = registry.get_recipe_instances(badges=badges, excluded=excluded)

    for instance in instances:
        reset_queries()
        instance.create_awards(
            batch_size=batch_size,
            db_read=db_read,
            post_save_signal=award_post_save)
        log_queries(instance)


def show_stats(**kwargs):
    """
    Shows badges stats.
    """
    db_read = kwargs.get('db_read', DEFAULT_DB_ALIAS)

    badges = (Badge.objects.using(db_read)
                           .all()
                           .annotate(u_count=Count('users'))
                           .order_by('u_count'))

    for badge in badges:
        logger.info('{:<20} {:>10} users awarded | users_count: {})'.format(
            badge.name,
            badge.u_count,
            badge.users_count))


def reset_awards(**kwargs):
    """
    Resets badges stats.
    """
    filter_badges = kwargs.get('badges', None)
    exclude_badges = kwargs.get('exclude_badges', None)

    for option in [filter_badges, exclude_badges]:
        if option:
            if not isinstance(option, (list, tuple)):
                option = [option]

    signals.pre_delete.disconnect(sender=Award,
                                  dispatch_uid='badgify.award.pre_delete.decrement_badge_users_count')

    award_qs = Award.objects.all()
    badge_qs = Badge.objects.all()

    if filter_badges:
        award_qs = award_qs.filter(badge__slug__in=filter_badges)
        badge_qs = badge_qs.filter(slug__in=filter_badges)

    if exclude_badges:
        award_qs = award_qs.exclude(badge__slug__in=exclude_badges)
        badge_qs = badge_qs.exclude(slug__in=exclude_badges)

    awards_count = award_qs.count()
    award_qs.delete()
    logger.info('✓ Deleted %d awards', awards_count)

    badges_count = badge_qs.count()
    badge_qs.update(users_count=0)
    logger.info('✓ Reseted Badge.users_count field of %d badge(s)', badges_count)



from django.db import models

from badgify import compat
from badgify.compat import gettext_lazy as _
from badgify.utils import get_model_string


class Award(models.Model):
    """
    An Award.
    """
    user = models.ForeignKey(
        compat.AUTH_USER_MODEL,
        verbose_name=_('user'),
        on_delete=models.CASCADE,
        related_name='badges')

    badge = models.ForeignKey(
        get_model_string('Badge'),
        verbose_name=_('badge'),
        on_delete=models.CASCADE,
        related_name='awards')

    awarded_at = models.DateTimeField(
        verbose_name=_('awarded at'),
        auto_now_add=True)

    class Meta:
        abstract = True
        app_label = 'badgify'
        verbose_name = _('award')
        verbose_name_plural = _('awards')
        unique_together = (('user', 'badge'),)

    def __str__(self):
        return '%s earned %s' % (self.user, self.badge.name)



# -*- coding: utf-8 -*-
import django

from .. import settings
from ..utils import load_class


Badge = load_class(settings.BADGE_MODEL)
Award = load_class(settings.AWARD_MODEL)

if django.VERSION < (1, 7):
    from .. import autodiscover
    autodiscover()

from ..signals import *   # noqa



import logging

logger = logging.getLogger('badgify')


class BadgifyRegistry(object):
    """
    Badge recipes registry.
    """

    def __init__(self):
        """
        Initializes a new registry.
        """
        self._registry = {}

    @property
    def recipes(self):
        """
        Returns all registered recipes (keys are recipe's badge slugs and values
        are recipe instances.
        """
        return self._registry

    @property
    def registered(self):
        """
        Returns registered recipes list (badge slugs).
        """
        return list(self._registry.keys())

    def register(self, recipe):
        """
        Registers a new recipe class.
        """

        if not isinstance(recipe, (list, tuple)):
            recipe = [recipe, ]

        for item in recipe:
            recipe = self.get_recipe_instance_from_class(item)
            self._registry[recipe.slug] = recipe

    def unregister(self, recipe):
        """
        Unregisters a given recipe class.
        """
        recipe = self.get_recipe_instance_from_class(recipe)
        if recipe.slug in self._registry:
            del self._registry[recipe.slug]

    def clear(self):
        """
        Clears the registry (removes all registered classes).
        """
        self._registry = {}

    def get_recipe_instance(self, badge):
        """
        Returns the recipe instance for the given badge slug.
        If badge has not been registered, raises ``exceptions.BadgeNotFound``.
        """
        from .exceptions import BadgeNotFound
        if badge in self._registry:
            return self.recipes[badge]
        raise BadgeNotFound()

    def get_recipe_instances(self, badges=None, excluded=None):
        """
        Returns all recipe instances or just those for the given badges.
        """
        if badges:
            if not isinstance(badges, (list, tuple)):
                badges = [badges]

        if excluded:
            if not isinstance(excluded, (list, tuple)):
                excluded = [excluded]
            badges = list(set(self.registered) - set(excluded))

        if badges:
            valid, invalid = self.get_recipe_instances_for_badges(badges=badges)
            return valid

        return self.recipes.values()

    def get_recipe_instances_for_badges(self, badges):
        """
        Takes a list of badge slugs and returns a tuple: ``(valid, invalid)``.
        """
        from .exceptions import BadgeNotFound

        valid, invalid = [], []

        if not isinstance(badges, (list, tuple)):
            badges = [badges]

        for badge in badges:
            try:
                recipe = self.get_recipe_instance(badge)
                valid.append(recipe)
            except BadgeNotFound:
                logger.debug('✘ Badge "%s" has not been registered', badge)
                invalid.append(badge)

        return (valid, invalid)

    @staticmethod
    def get_recipe_instance_from_class(klass):
        """
        Returns recipe instance from the given class ``klass``.
        """
        from .recipe import BaseRecipe
        assert issubclass(klass, BaseRecipe)
        return klass()


def _autodiscover(recipes):
    import copy
    from django.conf import settings

    try:
        # py27 / py3 only
        from importlib import import_module
    except ImportError:
        from django.utils.importlib import import_module

    from django.utils.module_loading import module_has_submodule

    for app in settings.INSTALLED_APPS:
        if app.startswith("badgify"):
            continue

        mod = import_module(app)
        try:
            before_import_recipes = copy.copy(recipes)
            import_module('%s.badgify_recipes' % app)
        except Exception:
            recipes = before_import_recipes
            if module_has_submodule(mod, 'badgify_recipes'):
                raise


registry = BadgifyRegistry()


def autodiscover():
    _autodiscover(registry)


def register(*args, **kwargs):
    return registry.register(*args, **kwargs)



# -*- coding: utf-8 -*-
from django.core.management.base import BaseCommand

from badgify.commands import reset_awards
from badgify.utils import sanitize_command_options


class Command(BaseCommand):
    """
    Commands that resets awards.
    """
    help = 'Resets awards'

    def add_arguments(self, parser):
        """
        Command arguments.
        """
        parser.add_argument('--badges',
                            action='store',
                            dest='badges',
                            type=str)

        parser.add_argument('--exclude-badges',
                            action='store',
                            dest='exclude_badges',
                            type=str)

    def handle(self, **options):
        """
        Command handler.
        """
        reset_awards(**sanitize_command_options(options))



import os

ROOT = os.path.dirname(os.path.abspath(__file__))

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": ":memory:",
    }
}

SITE_ID = 1

DEBUG = True

MIDDLEWARE = (
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
)

INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.sites",
    "django.contrib.messages",
    "badgify",
    "badgify.tests",
]

AUTH_USER_MODEL = "tests.BadgifyUser"

MEDIA_URL = "/media/"
MEDIA_ROOT = os.path.join(ROOT, "media")
STATIC_URL = "/static/"
STATIC_ROOT = os.path.join(ROOT, "static")

SECRET_KEY = "blabla"

ROOT_URLCONF = "badgify.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.contrib.auth.context_processors.auth",
                "django.template.context_processors.request",
                "django.template.context_processors.debug",
                "django.template.context_processors.i18n",
                "django.template.context_processors.media",
                "django.template.context_processors.static",
                "django.template.context_processors.tz",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "handlers": {
        "stream": {
            "level": "DEBUG",
            "class": "logging.StreamHandler",
        },
    },
    "loggers": {
        "django.request": {
            "handlers": ["stream"],
            "level": "ERROR",
            "propagate": True,
        },
        "badgify": {
            "handlers": ["stream"],
            "level": "ERROR",
            "propagate": True,
        },
    },
}

DEFAULT_AUTO_FIELD = "django.db.models.AutoField"
